@@ -341,8 +341,19 @@ static int cgi_demux_response(server *srv, handler_ctx *hctx) {
 
 	while(1) {
 		int n;
+		int toread;
+
+#if defined(__WIN32)
+		buffer_prepare_copy(hctx->response, 4 * 1024);
+#else
+		if (ioctl(con->fd, FIONREAD, &toread) || toread == 0 || toread <= 4*1024) {
+			buffer_prepare_copy(hctx->response, 4 * 1024);
+		} else {
+			if (toread > MAX_READ_LIMIT) toread = MAX_READ_LIMIT;
+			buffer_prepare_copy(hctx->response, toread + 1);
+		}
+#endif
 
-		buffer_prepare_copy(hctx->response, 1024);
 		if (-1 == (n = read(hctx->fd, hctx->response->ptr, hctx->response->size - 1))) {
 			if (errno == EAGAIN || errno == EINTR) {
 				/* would block, wait for signal */
@@ -577,6 +588,8 @@ static handler_t cgi_connection_close(server *srv, handler_ctx *hctx) {
 				connection_set_state(srv, con, CON_STATE_HANDLE_REQUEST);
 				con->http_status = 500;
 				con->mode = DIRECT;
+			} else {
+				con->file_finished = 1;
 			}
 
 			if (WIFEXITED(status)) {
@@ -607,8 +620,7 @@ static handler_t cgi_connection_close_callback(server *srv, connection *con, voi
 }
 
 
-static handler_t cgi_handle_fdevent(void *s, void *ctx, int revents) {
-	server      *srv  = (server *)s;
+static handler_t cgi_handle_fdevent(server *srv, void *ctx, int revents) {
 	handler_ctx *hctx = ctx;
 	connection  *con  = hctx->remote_conn;
 
@@ -635,9 +647,14 @@ static handler_t cgi_handle_fdevent(void *s, void *ctx, int revents) {
 			/* if we get a IN|HUP and have read everything don't exec the close twice */
 			return HANDLER_FINISHED;
 		case FDEVENT_HANDLED_ERROR:
-			connection_set_state(srv, con, CON_STATE_HANDLE_REQUEST);
-			con->http_status = 500;
-			con->mode = DIRECT;
+			/* Send an error if we haven't sent any data yet */
+			if (0 == con->file_started) {
+				connection_set_state(srv, con, CON_STATE_HANDLE_REQUEST);
+				con->http_status = 500;
+				con->mode = DIRECT;
+			} else {
+				con->file_finished = 1;
+			}
 
 			log_error_write(srv, __FILE__, __LINE__, "s", "demuxer failed: ");
 			break;
@@ -747,6 +764,8 @@ static int cgi_create_env(server *srv, connection *con, plugin_data *p, buffer *
 	}
 
 	if (pipe(from_cgi_fds)) {
+		close(to_cgi_fds[0]);
+		close(to_cgi_fds[1]);
 		log_error_write(srv, __FILE__, __LINE__, "ss", "pipe failed:", strerror(errno));
 		return -1;
 	}
@@ -909,7 +928,7 @@ static int cgi_create_env(server *srv, connection *con, plugin_data *p, buffer *
 		cgi_env_add(&env, CONST_STR_LEN("CONTENT_LENGTH"), buf, strlen(buf));
 		cgi_env_add(&env, CONST_STR_LEN("SCRIPT_FILENAME"), CONST_BUF_LEN(con->physical.path));
 		cgi_env_add(&env, CONST_STR_LEN("SCRIPT_NAME"), CONST_BUF_LEN(con->uri.path));
-		cgi_env_add(&env, CONST_STR_LEN("DOCUMENT_ROOT"), CONST_BUF_LEN(con->physical.doc_root));
+		cgi_env_add(&env, CONST_STR_LEN("DOCUMENT_ROOT"), CONST_BUF_LEN(con->physical.basedir));
 
 		/* for valgrind */
 		if (NULL != (s = getenv("LD_PRELOAD"))) {
@@ -1035,6 +1054,10 @@ static int cgi_create_env(server *srv, connection *con, plugin_data *p, buffer *
 	case -1:
 		/* error */
 		log_error_write(srv, __FILE__, __LINE__, "ss", "fork failed:", strerror(errno));
+		close(from_cgi_fds[0]);
+		close(from_cgi_fds[1]);
+		close(to_cgi_fds[0]);
+		close(to_cgi_fds[1]);
 		return -1;
 		break;
 	default: {
@@ -1145,7 +1168,7 @@ static int cgi_create_env(server *srv, connection *con, plugin_data *p, buffer *
 		con->plugin_ctx[p->id] = hctx;
 
 		fdevent_register(srv->ev, hctx->fd, cgi_handle_fdevent, hctx);
-		fdevent_event_add(srv->ev, &(hctx->fde_ndx), hctx->fd, FDEVENT_IN);
+		fdevent_event_set(srv->ev, &(hctx->fde_ndx), hctx->fd, FDEVENT_IN);
 
 		if (-1 == fdevent_fcntl_set(srv->ev, hctx->fd)) {
 			log_error_write(srv, __FILE__, __LINE__, "ss", "fcntl failed: ", strerror(errno));
@@ -1181,6 +1204,7 @@ static int mod_cgi_patch_connection(server *srv, connection *con, plugin_data *p
 	plugin_config *s = p->config_storage[0];
 
 	PATCH(cgi);
+	PATCH(execute_x_only);
 
 	/* skip the first, the global context */
 	for (i = 1; i < srv->config_context->used; i++) {
@@ -1264,6 +1288,15 @@ TRIGGER_FUNC(cgi_trigger) {
 #endif
 			break;
 		case -1:
+			if (errno == ECHILD) {
+				/* someone else called waitpid... remove the pid to stop looping the error each time */
+				log_error_write(srv, __FILE__, __LINE__, "s", "cgi child vanished, probably someone else called waitpid");
+
+				cgi_pid_del(srv, p, p->cgi_pid.ptr[ndx]);
+				ndx--;
+				continue;
+			}
+
 			log_error_write(srv, __FILE__, __LINE__, "ss", "waitpid failed: ", strerror(errno));
 
 			return HANDLER_ERROR;
@@ -1295,6 +1328,11 @@ TRIGGER_FUNC(cgi_trigger) {
 	return HANDLER_GO_ON;
 }
 
+/*
+ * - HANDLER_GO_ON : not our job
+ * - HANDLER_FINISHED: got response header
+ * - HANDLER_WAIT_FOR_EVENT: waiting for response header
+ */
 SUBREQUEST_FUNC(mod_cgi_handle_subrequest) {
 	int status;
 	plugin_data *p = p_d;
@@ -1306,7 +1344,13 @@ SUBREQUEST_FUNC(mod_cgi_handle_subrequest) {
 #if 0
 	log_error_write(srv, __FILE__, __LINE__, "sdd", "subrequest, pid =", hctx, hctx->pid);
 #endif
-	if (hctx->pid == 0) return HANDLER_FINISHED;
+
+	if (hctx->pid == 0) {
+		/* cgi already dead */
+		if (!con->file_started) return HANDLER_WAIT_FOR_EVENT;
+		return HANDLER_FINISHED;
+	}
+
 #ifndef __WIN32
 	switch(waitpid(hctx->pid, &status, WNOHANG)) {
 	case 0:
@@ -1344,24 +1388,24 @@ SUBREQUEST_FUNC(mod_cgi_handle_subrequest) {
 
 		return HANDLER_FINISHED;
 	default:
-		/* cgi process exited cleanly
-		 *
-		 * check if we already got the response
+		/* cgi process exited
 		 */
 
-		if (!con->file_started) return HANDLER_WAIT_FOR_EVENT;
+		hctx->pid = 0;
+
+		/* we already have response headers? just continue */
+		if (con->file_started) return HANDLER_FINISHED;
 
 		if (WIFEXITED(status)) {
-			/* nothing */
-		} else {
-			log_error_write(srv, __FILE__, __LINE__, "s", "cgi died ?");
-
-			con->mode = DIRECT;
-			con->http_status = 500;
-
+			/* clean exit - just continue */
+			return HANDLER_WAIT_FOR_EVENT;
 		}
 
-		hctx->pid = 0;
+		/* cgi proc died, and we didn't get any data yet - send error message and close cgi con */
+		log_error_write(srv, __FILE__, __LINE__, "s", "cgi died ?");
+
+		con->http_status = 500;
+		con->mode = DIRECT;
 
 		fdevent_event_del(srv->ev, &(hctx->fde_ndx), hctx->fd);
 		fdevent_unregister(srv->ev, hctx->fd);
